@@ -1,8 +1,7 @@
 import io
 import os
-import stat
 from os import PathLike
-from typing import Any, AnyStr, Dict, Mapping, Optional, Union
+from typing import Any, AnyStr, Mapping, Optional, Union
 
 from tqdm import tqdm
 
@@ -24,13 +23,13 @@ class Hasher:
     when calculating the hash value, a ``tqdm`` progress bar will be displayed.
 
     Parameters:
-        ctx (hash context):
+        ctx (hash-context):
             The hash context prototype used to generate hash values.
         chunksize (int | None, default=None):
             The chunk size for reading data from files.
         tqdm_args (Mapping | None, default=None):
             The arguments passed to the ``tqdm_class``.
-        tqdm_class (tqdm class | None, default=None):
+        tqdm_class (tqdm-class | None, default=None):
             The ``tqdm`` class.
     """
 
@@ -70,8 +69,8 @@ class Hasher:
             tqdm_class = tqdm
 
         self._ctx = ctx.copy()
-        self.chunksize: int = chunksize
-        self.tqdm_args: Dict[str, Any] = tqdm_args
+        self.chunksize = chunksize
+        self.tqdm_args = tqdm_args
         self.tqdm_class = tqdm_class
 
     def __call__(
@@ -85,7 +84,7 @@ class Hasher:
         """Return the hash value of a file or a directory.
 
         Parameters:
-            path (str, bytes or path-like):
+            path (AnyStr | PathLike[AnyStr]):
                 The path of a file or a directory.
             start (int | None, default=None):
                 The start offset of the file or files in the directory.
@@ -99,7 +98,7 @@ class Hasher:
                 If ``dir_ok`` is ``False`` and ``path`` is a directory.
 
         Returns:
-            bytes
+            bytes:
                 The hash value of the file or the directory.
         """
 
@@ -110,8 +109,7 @@ class Hasher:
             tn = type(stop).__name__
             raise TypeError(f"stop must be int or None, not {tn}")
 
-        st = os.stat(path)
-        if stat.S_ISDIR(st.st_mode):
+        if os.path.isdir(path):
             if dir_ok:
                 return self._hash_dir(path, start, stop)
             raise IsADirectory(f"{path!r} is a directory")
@@ -142,7 +140,7 @@ class Hasher:
         stop: Optional[int] = None,
     ) -> bytes:
         # Clamp ``(start, stop)`` to ``(0, filesize)``.
-        filesize = os.stat(filepath).st_size
+        filesize = os.path.getsize(filepath)
         if start is None or start < 0:
             start = 0
         if stop is None or stop > filesize:
